"""gustaf/gustaf/utils/log.py.

Thin logging wrapper.
"""

import logging


def configure(debug=False, logfile=None):
    """Logging configurator.

    Parameters
    -----------
    debug: bool
    logfile: str

    Returns
    --------
    None
    """
    logger = logging.getLogger()
    if debug:
        logger.setLevel(logging.DEBUG)

    else:
        logger.setLevel(logging.INFO)

    if logfile is not None:
        file_logger_handler = logging.FileHandler(logfile)
        logger.addHandler(file_logger_handler)
    logging.basicConfig(format="%(asctime)-15s [%(levelname)s]: %(message)s", )


def debug(*log):
    """Debug logger.

    Parameters
    -----------
    *log: Tuple[str]

    Returns
    --------
    None
    """
    logging.debug(" ".join(map(str, log)))


def info(*log):
    """Info logger.

    Parameters
    -----------
    *log: Tuple[str]

    Returns
    --------
    None
    """
    logging.info(" ".join(map(str, log)))


def warning(*log):
    """warning logger.

    Parameters
    -----------
    *log: Tuple[str]

    Returns
    --------
    None
    """
    logging.warning(" ".join(map(str, log)))



import numpy as np

from gustaf.spline import base


class InverseCrossTile3D(base.GustafBase):
    """Class that provides necessary functions to create inverse microtile,
    that can be used to describe the domain within a microstructure."""

    def __init__(self):
        """Simple inverse crosstile to tile with linear-quadratic branches and
        a trilinear center spline."""
        self._dim = 3
        self._evaluation_points = np.array(
                [
                        [0., .5, .5],
                        [1., .5, .5],
                        [.5, 0., .5],
                        [.5, 1., .5],
                        [.5, .5, 0.],
                        [.5, .5, 1.],
                ]
        )
        self._parameter_space_dimension = 1

    @property
    def parameter_space_dimension(self):
        """Number of parameters per evaluation point."""
        return self._parameter_space_dimension

    @property
    def evaluation_points(self):
        """Positions in the parametrization function to be evaluated when tile
        is constructed prior to composition.

        Parameters
        ----------
        None

        Returns
        -------
        evaluation_points : np.ndarray(6,3)
        """
        return self._evaluation_points

    @property
    def dim(self):
        """Returns dimensionality in physical space of the inverse tile.

        Parameters
        ----------
        None

        Returns
        -------
        dim : int
        """
        return self._dim

    def closing_tile(
            self,
            parameters=None,
            closure=None,
            boundary_width=0.1,
            filling_height=0.5,
            seperator_distance=None,
            **kwargs
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : tuple(np.ndarray)
          radii of fitting cylinder at evaluation points
        closure : str
          parametric dimension that needs to be closed.
          Must be one of {"z_min", "z_max"}
        boundary_width : float
          with of the boundary surronding branch
        filling_height : float
          portion of the height that is filled in parametric domain

        Results
        -------
        list_of_splines : list
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        if seperator_distance is None:
            raise ValueError(
                    "Seperator Distance is missing. The value is required to "
                    "create watertight connections with neighboring elements."
                    " The value should be greater than the biggest branch "
                    "radius"
            )

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        parameters = parameters[0]
        if not (np.all(parameters > 0) and np.all(parameters < .5)):
            raise ValueError("Thickness out of range (0, .5)")

        if not (0. < float(boundary_width) < .5):
            raise ValueError("Boundary Width is out of range")

        if not (0. < float(filling_height) < 1.):
            raise ValueError("Filling must  be in (0,1)")

        # Precompute auxiliary values
        inv_filling_height = 1. - filling_height
        ctps_mid_height_top = (1 + filling_height) * .5
        ctps_mid_height_bottom = 1. - ctps_mid_height_top
        center_width = 1. - 2 * boundary_width
        r_center = center_width * .5
        half_r_center = (r_center + .5) * .5
        aux_column_width = .5 - 2 * (.5 - seperator_distance)

        spline_list = []
        if closure == "z_min":
            branch_thickness = parameters[5]
            branch_neighbor_x_min_ctps = np.array(
                    [
                            [-.5, -r_center, filling_height],
                            [-half_r_center, -r_center, filling_height],
                            [-r_center, -r_center, filling_height],
                            [-.5, r_center, filling_height],
                            [-half_r_center, r_center, filling_height],
                            [-r_center, r_center, filling_height],
                            [-.5, -aux_column_width, ctps_mid_height_top],
                            [
                                    -seperator_distance, -aux_column_width,
                                    ctps_mid_height_top
                            ],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ], [-.5, aux_column_width, ctps_mid_height_top],
                            [
                                    -seperator_distance, aux_column_width,
                                    ctps_mid_height_top
                            ],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ], [-.5, -aux_column_width, 1.],
                            [-seperator_distance, -aux_column_width, 1.],
                            [-branch_thickness, -branch_thickness, 1.],
                            [-.5, aux_column_width, 1.],
                            [-seperator_distance, aux_column_width, 1.],
                            [-branch_thickness, branch_thickness, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 1, 2],
                            control_points=branch_neighbor_x_min_ctps
                    )
            )

            branch_neighbor_x_max_ctps = np.array(
                    [
                            [r_center, -r_center, filling_height],
                            [half_r_center, -r_center, filling_height],
                            [.5, -r_center, filling_height],
                            [r_center, r_center, filling_height],
                            [half_r_center, r_center, filling_height],
                            [.5, r_center, filling_height],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    seperator_distance, -aux_column_width,
                                    ctps_mid_height_top
                            ], [.5, -aux_column_width, ctps_mid_height_top],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    seperator_distance, aux_column_width,
                                    ctps_mid_height_top
                            ], [.5, aux_column_width, ctps_mid_height_top],
                            [branch_thickness, -branch_thickness, 1.],
                            [seperator_distance, -aux_column_width, 1.],
                            [.5, -aux_column_width, 1.],
                            [branch_thickness, branch_thickness, 1.],
                            [seperator_distance, aux_column_width, 1.],
                            [.5, aux_column_width, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 1, 2],
                            control_points=branch_neighbor_x_max_ctps
                    )
            )

            branch_neighbor_y_min_ctps = np.array(
                    [
                            [-r_center, -.5, filling_height],
                            [r_center, -.5, filling_height],
                            [-r_center, -half_r_center, filling_height],
                            [r_center, -half_r_center, filling_height],
                            [-r_center, -r_center, filling_height],
                            [r_center, -r_center, filling_height],
                            [-aux_column_width, -.5, ctps_mid_height_top],
                            [aux_column_width, -.5, ctps_mid_height_top],
                            [
                                    -aux_column_width, -seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [
                                    aux_column_width, -seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ], [-aux_column_width, -.5, 1.],
                            [aux_column_width, -.5, 1.],
                            [-aux_column_width, -seperator_distance, 1.],
                            [aux_column_width, -seperator_distance, 1.],
                            [-branch_thickness, -branch_thickness, 1.],
                            [branch_thickness, -branch_thickness, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 2, 2],
                            control_points=branch_neighbor_y_min_ctps
                    )
            )

            branch_neighbor_y_max_ctps = np.array(
                    [
                            [-r_center, r_center, filling_height],
                            [r_center, r_center, filling_height],
                            [-r_center, half_r_center, filling_height],
                            [r_center, half_r_center, filling_height],
                            [-r_center, .5, filling_height],
                            [r_center, .5, filling_height],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    -aux_column_width, seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [
                                    aux_column_width, seperator_distance,
                                    ctps_mid_height_top
                            ], [-aux_column_width, .5, ctps_mid_height_top],
                            [aux_column_width, .5, ctps_mid_height_top],
                            [-branch_thickness, branch_thickness, 1.],
                            [branch_thickness, branch_thickness, 1.],
                            [-aux_column_width, seperator_distance, 1.],
                            [aux_column_width, seperator_distance, 1.],
                            [-aux_column_width, .5, 1.],
                            [aux_column_width, .5, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 2, 2],
                            control_points=branch_neighbor_y_max_ctps
                    )
            )

            branch_x_min_y_min_ctps = np.array(
                    [
                            [-.5, -.5, filling_height],
                            [-half_r_center, -.5, filling_height],
                            [-r_center, -.5, filling_height],
                            [-.5, -half_r_center, filling_height],
                            [-half_r_center, -half_r_center, filling_height],
                            [-r_center, -half_r_center, filling_height],
                            [-.5, -r_center, filling_height],
                            [-half_r_center, -r_center, filling_height],
                            [-r_center, -r_center, filling_height],
                            [-.5, -.5, ctps_mid_height_top],
                            [-seperator_distance, -.5, ctps_mid_height_top],
                            [-aux_column_width, -.5, ctps_mid_height_top],
                            [-.5, -seperator_distance, ctps_mid_height_top],
                            [
                                    -seperator_distance, -seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [
                                    -aux_column_width, -seperator_distance,
                                    ctps_mid_height_top
                            ], [-.5, -aux_column_width, ctps_mid_height_top],
                            [
                                    -seperator_distance, -aux_column_width,
                                    ctps_mid_height_top
                            ],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ], [-.5, -.5, 1.], [-seperator_distance, -.5, 1.],
                            [-aux_column_width, -.5, 1.],
                            [-.5, -seperator_distance, 1.],
                            [-seperator_distance, -seperator_distance, 1.],
                            [-aux_column_width, -seperator_distance, 1.],
                            [-.5, -aux_column_width, 1.],
                            [-seperator_distance, -aux_column_width, 1.],
                            [-branch_thickness, -branch_thickness, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_min_y_min_ctps
                    )
            )

            branch_x_min_y_max_ctps = np.array(
                    [
                            [-.5, r_center, filling_height],
                            [-half_r_center, r_center, filling_height],
                            [-r_center, r_center, filling_height],
                            [-.5, half_r_center, filling_height],
                            [-half_r_center, half_r_center, filling_height],
                            [-r_center, half_r_center, filling_height],
                            [-.5, .5, filling_height],
                            [-half_r_center, .5, filling_height],
                            [-r_center, .5, filling_height],
                            [-.5, aux_column_width, ctps_mid_height_top],
                            [
                                    -seperator_distance, aux_column_width,
                                    ctps_mid_height_top
                            ],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ], [-.5, seperator_distance, ctps_mid_height_top],
                            [
                                    -seperator_distance, seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [
                                    -aux_column_width, seperator_distance,
                                    ctps_mid_height_top
                            ], [-.5, .5, ctps_mid_height_top],
                            [-seperator_distance, .5, ctps_mid_height_top],
                            [-aux_column_width, .5, ctps_mid_height_top],
                            [-.5, aux_column_width, 1.],
                            [-seperator_distance, aux_column_width, 1.],
                            [-branch_thickness, branch_thickness, 1.],
                            [-.5, seperator_distance, 1.],
                            [-seperator_distance, seperator_distance, 1.],
                            [-aux_column_width, seperator_distance, 1.],
                            [-.5, .5, 1.], [-seperator_distance, .5, 1.],
                            [-aux_column_width, .5, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_min_y_max_ctps
                    )
            )

            branch_x_max_y_min_ctps = np.array(
                    [
                            [r_center, -.5, filling_height],
                            [half_r_center, -.5, filling_height],
                            [.5, -.5, filling_height],
                            [r_center, -half_r_center, filling_height],
                            [half_r_center, -half_r_center, filling_height],
                            [.5, -half_r_center, filling_height],
                            [r_center, -r_center, filling_height],
                            [half_r_center, -r_center, filling_height],
                            [.5, -r_center, filling_height],
                            [aux_column_width, -.5, ctps_mid_height_top],
                            [seperator_distance, -.5, ctps_mid_height_top],
                            [.5, -.5, ctps_mid_height_top],
                            [
                                    aux_column_width, -seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [
                                    seperator_distance, -seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [.5, -seperator_distance, ctps_mid_height_top],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    seperator_distance, -aux_column_width,
                                    ctps_mid_height_top
                            ],
                            [.5, -aux_column_width, ctps_mid_height_top],
                            [aux_column_width, -.5, 1.],
                            [seperator_distance, -.5, 1.],
                            [.5, -.5, 1.],
                            [aux_column_width, -seperator_distance, 1.],
                            [seperator_distance, -seperator_distance, 1.],
                            [.5, -seperator_distance, 1.],
                            [branch_thickness, -branch_thickness, 1.],
                            [seperator_distance, -aux_column_width, 1.],
                            [.5, -aux_column_width, 1.],
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_max_y_min_ctps
                    )
            )

            branch_x_max_y_max_ctps = np.array(
                    [
                            [r_center, r_center, filling_height],
                            [half_r_center, r_center, filling_height],
                            [.5, r_center, filling_height],
                            [r_center, half_r_center, filling_height],
                            [half_r_center, half_r_center, filling_height],
                            [.5, half_r_center, filling_height],
                            [r_center, .5, filling_height],
                            [half_r_center, .5, filling_height],
                            [.5, .5, filling_height],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_top
                            ],
                            [
                                    seperator_distance, aux_column_width,
                                    ctps_mid_height_top
                            ], [.5, aux_column_width, ctps_mid_height_top],
                            [
                                    aux_column_width, seperator_distance,
                                    ctps_mid_height_top
                            ],
                            [
                                    seperator_distance, seperator_distance,
                                    ctps_mid_height_top
                            ], [.5, seperator_distance, ctps_mid_height_top],
                            [aux_column_width, .5, ctps_mid_height_top],
                            [seperator_distance, .5, ctps_mid_height_top],
                            [.5, .5, ctps_mid_height_top],
                            [branch_thickness, branch_thickness, 1.],
                            [seperator_distance, aux_column_width, 1.],
                            [.5, aux_column_width, 1.],
                            [aux_column_width, seperator_distance, 1.],
                            [seperator_distance, seperator_distance, 1.],
                            [.5, seperator_distance, 1.],
                            [aux_column_width, .5, 1.],
                            [seperator_distance, .5, 1.], [.5, .5, 1.]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_max_y_max_ctps
                    )
            )

            return spline_list

        elif closure == "z_max":
            branch_thickness = parameters[4]
            branch_neighbor_x_min_ctps = np.array(
                    [
                            [-.5, -aux_column_width, 0.],
                            [-seperator_distance, -aux_column_width, 0.],
                            [-branch_thickness, -branch_thickness, 0.],
                            [-.5, aux_column_width, 0.],
                            [-seperator_distance, aux_column_width, 0.],
                            [-branch_thickness, branch_thickness, 0.],
                            [-.5, -aux_column_width, ctps_mid_height_bottom],
                            [
                                    -seperator_distance, -aux_column_width,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ], [-.5, aux_column_width, ctps_mid_height_bottom],
                            [
                                    -seperator_distance, aux_column_width,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ], [-.5, -r_center, inv_filling_height],
                            [-half_r_center, -r_center, inv_filling_height],
                            [-r_center, -r_center, inv_filling_height],
                            [-.5, r_center, inv_filling_height],
                            [-half_r_center, r_center, inv_filling_height],
                            [-r_center, r_center, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 1, 2],
                            control_points=branch_neighbor_x_min_ctps
                    )
            )

            branch_neighbor_x_max_ctps = np.array(
                    [
                            [branch_thickness, -branch_thickness, 0.],
                            [seperator_distance, -aux_column_width, 0.],
                            [.5, -aux_column_width, 0.],
                            [branch_thickness, branch_thickness, 0.],
                            [seperator_distance, aux_column_width, 0.],
                            [.5, aux_column_width, 0.],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    seperator_distance, -aux_column_width,
                                    ctps_mid_height_bottom
                            ], [.5, -aux_column_width, ctps_mid_height_bottom],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    seperator_distance, aux_column_width,
                                    ctps_mid_height_bottom
                            ], [.5, aux_column_width, ctps_mid_height_bottom],
                            [r_center, -r_center, inv_filling_height],
                            [half_r_center, -r_center, inv_filling_height],
                            [.5, -r_center, inv_filling_height],
                            [r_center, r_center, inv_filling_height],
                            [half_r_center, r_center, inv_filling_height],
                            [.5, r_center, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 1, 2],
                            control_points=branch_neighbor_x_max_ctps
                    )
            )

            branch_neighbor_y_min_ctps = np.array(
                    [
                            [-aux_column_width, -.5, 0.],
                            [aux_column_width, -.5, 0.],
                            [-aux_column_width, -seperator_distance, 0.],
                            [aux_column_width, -seperator_distance, 0.],
                            [-branch_thickness, -branch_thickness, 0.],
                            [branch_thickness, -branch_thickness, 0.],
                            [-aux_column_width, -.5, ctps_mid_height_bottom],
                            [aux_column_width, -.5, ctps_mid_height_bottom],
                            [
                                    -aux_column_width, -seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    aux_column_width, -seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ], [-r_center, -.5, inv_filling_height],
                            [r_center, -.5, inv_filling_height],
                            [-r_center, -half_r_center, inv_filling_height],
                            [r_center, -half_r_center, inv_filling_height],
                            [-r_center, -r_center, inv_filling_height],
                            [r_center, -r_center, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 2, 2],
                            control_points=branch_neighbor_y_min_ctps
                    )
            )

            branch_neighbor_y_max_ctps = np.array(
                    [
                            [-branch_thickness, branch_thickness, 0.],
                            [branch_thickness, branch_thickness, 0.],
                            [-aux_column_width, seperator_distance, 0.],
                            [aux_column_width, seperator_distance, 0.],
                            [-aux_column_width, .5, 0.],
                            [aux_column_width, .5, 0.],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -aux_column_width, seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    aux_column_width, seperator_distance,
                                    ctps_mid_height_bottom
                            ], [-aux_column_width, .5, ctps_mid_height_bottom],
                            [aux_column_width, .5, ctps_mid_height_bottom],
                            [-r_center, r_center, inv_filling_height],
                            [r_center, r_center, inv_filling_height],
                            [-r_center, half_r_center, inv_filling_height],
                            [r_center, half_r_center, inv_filling_height],
                            [-r_center, .5, inv_filling_height],
                            [r_center, .5, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[1, 2, 2],
                            control_points=branch_neighbor_y_max_ctps
                    )
            )

            branch_x_min_y_min_ctps = np.array(
                    [
                            [-.5, -.5, 0.], [-seperator_distance, -.5, 0.],
                            [-aux_column_width, -.5, 0.],
                            [-.5, -seperator_distance, 0.],
                            [-seperator_distance, -seperator_distance, 0.],
                            [-aux_column_width, -seperator_distance, 0.],
                            [-.5, -aux_column_width, 0.],
                            [-seperator_distance, -aux_column_width, 0.],
                            [-branch_thickness, -branch_thickness, 0.],
                            [-.5, -.5, ctps_mid_height_bottom],
                            [-seperator_distance, -.5, ctps_mid_height_bottom],
                            [-aux_column_width, -.5, ctps_mid_height_bottom],
                            [-.5, -seperator_distance, ctps_mid_height_bottom],
                            [
                                    -seperator_distance, -seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -aux_column_width, -seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [-.5, -aux_column_width, ctps_mid_height_bottom],
                            [
                                    -seperator_distance, -aux_column_width,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ], [-.5, -.5, inv_filling_height],
                            [-half_r_center, -.5, inv_filling_height],
                            [-r_center, -.5, inv_filling_height],
                            [-.5, -half_r_center, inv_filling_height],
                            [
                                    -half_r_center, -half_r_center,
                                    inv_filling_height
                            ], [-r_center, -half_r_center, inv_filling_height],
                            [-.5, -r_center, inv_filling_height],
                            [-half_r_center, -r_center, inv_filling_height],
                            [-r_center, -r_center, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_min_y_min_ctps
                    )
            )

            branch_x_max_y_max_ctps = np.array(
                    [
                            [branch_thickness, branch_thickness, 0.],
                            [seperator_distance, aux_column_width, 0.],
                            [.5, aux_column_width, 0.],
                            [aux_column_width, seperator_distance, 0.],
                            [seperator_distance, seperator_distance, 0.],
                            [.5, seperator_distance, 0.],
                            [aux_column_width, .5, 0.],
                            [seperator_distance, .5, 0.], [.5, .5, 0.],
                            [
                                    branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    seperator_distance, aux_column_width,
                                    ctps_mid_height_bottom
                            ], [.5, aux_column_width, ctps_mid_height_bottom],
                            [
                                    aux_column_width, seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    seperator_distance, seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [.5, seperator_distance, ctps_mid_height_bottom],
                            [aux_column_width, .5, ctps_mid_height_bottom],
                            [seperator_distance, .5, ctps_mid_height_bottom],
                            [.5, .5, ctps_mid_height_bottom],
                            [r_center, r_center, inv_filling_height],
                            [half_r_center, r_center, inv_filling_height],
                            [.5, r_center, inv_filling_height],
                            [r_center, half_r_center, inv_filling_height],
                            [half_r_center, half_r_center, inv_filling_height],
                            [.5, half_r_center, inv_filling_height],
                            [r_center, .5, inv_filling_height],
                            [half_r_center, .5, inv_filling_height],
                            [.5, .5, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_max_y_max_ctps
                    )
            )

            branch_x_max_y_min_ctps = np.array(
                    [
                            [aux_column_width, -.5, 0.],
                            [seperator_distance, -.5, 0.], [.5, -.5, 0.],
                            [aux_column_width, -seperator_distance, 0.],
                            [seperator_distance, -seperator_distance, 0.],
                            [.5, -seperator_distance, 0.],
                            [branch_thickness, -branch_thickness, 0.],
                            [seperator_distance, -aux_column_width, 0.],
                            [.5, -aux_column_width, 0.],
                            [aux_column_width, -.5, ctps_mid_height_bottom],
                            [seperator_distance, -.5, ctps_mid_height_bottom],
                            [.5, -.5, ctps_mid_height_bottom],
                            [
                                    aux_column_width, -seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    seperator_distance, -seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [.5, -seperator_distance, ctps_mid_height_bottom],
                            [
                                    branch_thickness, -branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    seperator_distance, -aux_column_width,
                                    ctps_mid_height_bottom
                            ], [.5, -aux_column_width, ctps_mid_height_bottom],
                            [r_center, -.5, inv_filling_height],
                            [half_r_center, -.5, inv_filling_height],
                            [.5, -.5, inv_filling_height],
                            [r_center, -half_r_center, inv_filling_height],
                            [
                                    half_r_center, -half_r_center,
                                    inv_filling_height
                            ], [.5, -half_r_center, inv_filling_height],
                            [r_center, -r_center, inv_filling_height],
                            [half_r_center, -r_center, inv_filling_height],
                            [.5, -r_center, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_max_y_min_ctps
                    )
            )

            branch_x_min_y_max_ctps = np.array(
                    [
                            [-.5, aux_column_width, 0.],
                            [-seperator_distance, aux_column_width, 0.],
                            [-branch_thickness, branch_thickness, 0.],
                            [-.5, seperator_distance, 0.],
                            [-seperator_distance, seperator_distance, 0.],
                            [-aux_column_width, seperator_distance, 0.],
                            [-.5, .5, 0.], [-seperator_distance, .5, 0.],
                            [-aux_column_width, .5, 0.],
                            [-.5, aux_column_width, ctps_mid_height_bottom],
                            [
                                    -seperator_distance, aux_column_width,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -branch_thickness, branch_thickness,
                                    ctps_mid_height_bottom
                            ],
                            [-.5, seperator_distance, ctps_mid_height_bottom],
                            [
                                    -seperator_distance, seperator_distance,
                                    ctps_mid_height_bottom
                            ],
                            [
                                    -aux_column_width, seperator_distance,
                                    ctps_mid_height_bottom
                            ], [-.5, .5, ctps_mid_height_bottom],
                            [-seperator_distance, .5, ctps_mid_height_bottom],
                            [-aux_column_width, .5, ctps_mid_height_bottom],
                            [-.5, r_center, inv_filling_height],
                            [-half_r_center, r_center, inv_filling_height],
                            [-r_center, r_center, inv_filling_height],
                            [-.5, half_r_center, inv_filling_height],
                            [
                                    -half_r_center, half_r_center,
                                    inv_filling_height
                            ], [-r_center, half_r_center, inv_filling_height],
                            [-.5, .5, inv_filling_height],
                            [-half_r_center, .5, inv_filling_height],
                            [-r_center, .5, inv_filling_height]
                    ]
            ) + np.array([.5, .5, 0.])

            spline_list.append(
                    base.Bezier(
                            degrees=[2, 2, 2],
                            control_points=branch_x_min_y_max_ctps
                    )
            )

            return spline_list
        else:
            raise ValueError("Corner Type not supported")

    def create_tile(
            self,
            parameters=None,
            seperator_distance=None,
            center_expansion=1.,
            **kwargs
    ):
        """Create an inverse microtile based on the parameters that describe
        the branch thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : tuple(np.array)
          only first entry is used, defines the internal radii of the
          branches
        seperator_distance : float
          position of the control points for higher order elements
        center_expansion : float
          thickness of center is expanded by a factor

        Returns
        -------
        microtile_list : list(splines)
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")
        if not ((center_expansion > .5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5, 1.5)")

        if seperator_distance is None:
            raise ValueError(
                    "Seperator Distance is missing. The value is required to "
                    "create watertight connections with neighboring elements."
                    " The value should be greater than the biggest branch "
                    "radius"
            )

        # Check if all radii are in allowed range
        max_radius = min(.5, (.5 / center_expansion))
        max_radius = min(max_radius, seperator_distance)

        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r,
         z_max_r] = parameters[0].tolist()

        for radius in [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r]:
            if not isinstance(radius, float):
                raise ValueError("Invalid type")
            if not (radius > 0 and radius < max_radius):
                raise ValueError(
                        f"Radii must be in (0,{max_radius}) for "
                        f"center_expansion {center_expansion}"
                )

        # center radius
        center_r = (
                x_min_r + x_max_r + y_min_r + y_max_r + z_min_r + z_max_r
        ) / 6. * center_expansion

        # Auxiliary values for smooothing (mid-branch thickness)
        aux_x_min = min(x_min_r, center_r)
        aux_x_max = min(x_max_r, center_r)
        aux_y_min = min(y_min_r, center_r)
        aux_y_max = min(y_max_r, center_r)
        aux_z_min = min(z_min_r, center_r)
        aux_z_max = min(z_max_r, center_r)
        # Branch midlength
        hd_center = 0.5 * (0.5 + center_r)

        #
        if seperator_distance is None:
            seperator_distance = .45
        aux_column_width = .5 - 2 * (.5 - seperator_distance)

        # Init return type
        spline_list = []

        # Start with branch interconnections
        x_min_y_min = np.array(
                [
                        [-.5, -.5, -aux_column_width],
                        [-seperator_distance, -.5, -aux_column_width],
                        [-y_min_r, -.5, -y_min_r],
                        [-.5, -seperator_distance, -aux_column_width],
                        [-hd_center, -hd_center, -aux_column_width],
                        [-aux_y_min, -hd_center, -aux_y_min],
                        [-.5, -x_min_r, -x_min_r],
                        [-hd_center, -aux_x_min, -aux_x_min],
                        [-center_r, -center_r, -center_r],
                        [-.5, -.5, aux_column_width],
                        [-seperator_distance, -.5, aux_column_width],
                        [-y_min_r, -.5, y_min_r],
                        [-.5, -seperator_distance, aux_column_width],
                        [-hd_center, -hd_center, aux_column_width],
                        [-aux_y_min, -hd_center, aux_y_min],
                        [-.5, -x_min_r, x_min_r],
                        [-hd_center, -aux_x_min, aux_x_min],
                        [-center_r, -center_r, center_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 2, 1], control_points=x_min_y_min)
        )

        x_max_y_min = np.array(
                [
                        [y_min_r, -.5, -y_min_r],
                        [seperator_distance, -.5, -aux_column_width],
                        [.5, -.5, -aux_column_width],
                        [aux_y_min, -hd_center, -aux_y_min],
                        [hd_center, -hd_center, -aux_column_width],
                        [.5, -seperator_distance, -aux_column_width],
                        [center_r, -center_r, -center_r],
                        [hd_center, -aux_x_max, -aux_x_max],
                        [.5, -x_max_r, -x_max_r], [y_min_r, -.5, y_min_r],
                        [seperator_distance, -.5, aux_column_width],
                        [.5, -.5, aux_column_width],
                        [aux_y_min, -hd_center, aux_y_min],
                        [hd_center, -hd_center, aux_column_width],
                        [.5, -seperator_distance, aux_column_width],
                        [center_r, -center_r, center_r],
                        [hd_center, -aux_x_max, aux_x_max],
                        [.5, -x_max_r, x_max_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 2, 1], control_points=x_max_y_min)
        )

        x_min_y_max = np.array(
                [
                        [-.5, x_min_r, -x_min_r],
                        [-hd_center, aux_x_min, -aux_x_min],
                        [-center_r, center_r, -center_r],
                        [-.5, seperator_distance, -aux_column_width],
                        [-hd_center, hd_center, -aux_column_width],
                        [-aux_y_max, hd_center, -aux_y_max],
                        [-.5, .5, -aux_column_width],
                        [-seperator_distance, .5, -aux_column_width],
                        [-y_max_r, .5, -y_max_r], [-.5, x_min_r, x_min_r],
                        [-hd_center, aux_x_min, aux_x_min],
                        [-center_r, center_r, center_r],
                        [-.5, seperator_distance, aux_column_width],
                        [-hd_center, hd_center, aux_column_width],
                        [-aux_y_max, hd_center, aux_y_max],
                        [-.5, .5, aux_column_width],
                        [-seperator_distance, .5, aux_column_width],
                        [-y_max_r, .5, y_max_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 2, 1], control_points=x_min_y_max)
        )

        x_max_y_max = np.array(
                [
                        [center_r, center_r, -center_r],
                        [hd_center, aux_x_max, -aux_x_max],
                        [.5, x_max_r, -x_max_r],
                        [aux_y_max, hd_center, -aux_y_max],
                        [hd_center, hd_center, -aux_column_width],
                        [.5, seperator_distance, -aux_column_width],
                        [y_max_r, .5, -y_max_r],
                        [seperator_distance, .5, -aux_column_width],
                        [.5, .5, -aux_column_width],
                        [center_r, center_r, center_r],
                        [hd_center, aux_x_max, aux_x_max],
                        [.5, x_max_r, x_max_r],
                        [aux_y_max, hd_center, aux_y_max],
                        [hd_center, hd_center, aux_column_width],
                        [.5, seperator_distance, aux_column_width],
                        [y_max_r, .5, y_max_r],
                        [seperator_distance, .5, aux_column_width],
                        [.5, .5, aux_column_width]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 2, 1], control_points=x_max_y_max)
        )

        x_min_z_min = np.array(
                [
                        [-.5, -aux_column_width, -.5],
                        [-seperator_distance, -aux_column_width, -.5],
                        [-z_min_r, -z_min_r, -.5], [
                                -.5, aux_column_width, -.5
                        ], [-seperator_distance, aux_column_width, -.5],
                        [-z_min_r, z_min_r, -.5],
                        [-.5, -aux_column_width, -seperator_distance],
                        [-hd_center, -aux_column_width, -hd_center],
                        [-aux_z_min, -aux_z_min, -hd_center],
                        [-.5, aux_column_width, -seperator_distance],
                        [-hd_center, aux_column_width, -hd_center],
                        [-aux_z_min, aux_z_min, -hd_center],
                        [-.5, -x_min_r, -x_min_r],
                        [-hd_center, -aux_x_min, -aux_x_min],
                        [-center_r, -center_r, -center_r],
                        [-.5, x_min_r, -x_min_r],
                        [-hd_center, aux_x_min, -aux_x_min],
                        [-center_r, center_r, -center_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 1, 2], control_points=x_min_z_min)
        )

        x_max_z_min = np.array(
                [
                        [z_min_r, -z_min_r, -.5],
                        [seperator_distance, -aux_column_width, -.5],
                        [.5, -aux_column_width, -.5], [z_min_r, z_min_r, -.5],
                        [seperator_distance, aux_column_width, -.5],
                        [.5, aux_column_width, -.5],
                        [aux_z_min, -aux_z_min, -hd_center],
                        [hd_center, -aux_column_width, -hd_center],
                        [.5, -aux_column_width, -seperator_distance],
                        [aux_z_min, aux_z_min, -hd_center],
                        [hd_center, aux_column_width, -hd_center],
                        [.5, aux_column_width, -seperator_distance],
                        [center_r, -center_r, -center_r],
                        [hd_center, -aux_x_max, -aux_x_max],
                        [.5, -x_max_r, -x_max_r],
                        [center_r, center_r, -center_r],
                        [hd_center, aux_x_max, -aux_x_max],
                        [.5, x_max_r, -x_max_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 1, 2], control_points=x_max_z_min)
        )

        x_min_z_max = np.array(
                [
                        [-.5, -x_min_r, x_min_r],
                        [-hd_center, -aux_x_min, aux_x_min],
                        [-center_r, -center_r, center_r],
                        [-.5, x_min_r, x_min_r],
                        [-hd_center, aux_x_min, aux_x_min],
                        [-center_r, center_r, center_r],
                        [-.5, -aux_column_width, seperator_distance],
                        [-hd_center, -aux_column_width, hd_center],
                        [-aux_z_max, -aux_z_max, hd_center],
                        [-.5, aux_column_width, seperator_distance],
                        [-hd_center, aux_column_width, hd_center],
                        [-aux_z_max, aux_z_max, hd_center],
                        [-.5, -aux_column_width, .5],
                        [-seperator_distance, -aux_column_width, .5],
                        [-z_max_r, -z_max_r, .5], [-.5, aux_column_width, .5],
                        [-seperator_distance, aux_column_width, .5],
                        [-z_max_r, z_max_r, .5]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 1, 2], control_points=x_min_z_max)
        )

        x_max_z_max = np.array(
                [
                        [center_r, -center_r, center_r],
                        [hd_center, -aux_x_max, aux_x_max],
                        [.5, -x_max_r, x_max_r], [
                                center_r, center_r, center_r
                        ], [hd_center, aux_x_max, aux_x_max],
                        [.5, x_max_r, x_max_r],
                        [aux_z_max, -aux_z_max, hd_center],
                        [hd_center, -aux_column_width, hd_center],
                        [.5, -aux_column_width, seperator_distance],
                        [aux_z_max, aux_z_max, hd_center],
                        [hd_center, aux_column_width, hd_center],
                        [.5, aux_column_width, seperator_distance],
                        [z_max_r, -z_max_r, .5],
                        [seperator_distance, -aux_column_width, .5],
                        [.5, -aux_column_width, .5], [z_max_r, z_max_r, .5],
                        [seperator_distance, aux_column_width, .5],
                        [.5, aux_column_width, .5]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[2, 1, 2], control_points=x_max_z_max)
        )

        y_min_z_min = np.array(
                [
                        [-aux_column_width, -.5, -.5],
                        [aux_column_width, -.5, -.5],
                        [-aux_column_width, -seperator_distance, -.5],
                        [aux_column_width, -seperator_distance, -.5],
                        [-z_min_r, -z_min_r, -.5], [z_min_r, -z_min_r, -.5],
                        [-aux_column_width, -.5, -seperator_distance],
                        [aux_column_width, -.5, -seperator_distance],
                        [-aux_column_width, -hd_center, -hd_center],
                        [aux_column_width, -hd_center, -hd_center],
                        [-aux_z_min, -aux_z_min, -hd_center],
                        [aux_z_min, -aux_z_min, -hd_center],
                        [-y_min_r, -.5, -y_min_r], [y_min_r, -.5, -y_min_r],
                        [-aux_y_min, -hd_center, -aux_y_min],
                        [aux_y_min, -hd_center, -aux_y_min],
                        [-center_r, -center_r, -center_r],
                        [center_r, -center_r, -center_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[1, 2, 2], control_points=y_min_z_min)
        )

        y_max_z_min = np.array(
                [
                        [-z_min_r, z_min_r, -.5], [z_min_r, z_min_r, -.5],
                        [-aux_column_width, seperator_distance, -.5],
                        [aux_column_width, seperator_distance, -.5],
                        [-aux_column_width, .5, -.5],
                        [aux_column_width, .5, -.5],
                        [-aux_z_min, aux_z_min, -hd_center],
                        [aux_z_min, aux_z_min, -hd_center],
                        [-aux_column_width, hd_center, -hd_center],
                        [aux_column_width, hd_center, -hd_center],
                        [-aux_column_width, .5, -seperator_distance],
                        [aux_column_width, .5, -seperator_distance],
                        [-center_r, center_r, -center_r],
                        [center_r, center_r, -center_r],
                        [-aux_y_max, hd_center, -aux_y_max],
                        [aux_y_max, hd_center, -aux_y_max],
                        [-y_max_r, .5, -y_max_r], [y_max_r, .5, -y_max_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[1, 2, 2], control_points=y_max_z_min)
        )

        y_min_z_max = np.array(
                [
                        [-y_min_r, -.5, y_min_r],
                        [y_min_r, -.5, y_min_r],
                        [-aux_y_min, -hd_center, aux_y_min],
                        [aux_y_min, -hd_center, aux_y_min],
                        [-center_r, -center_r, center_r],
                        [center_r, -center_r, center_r],
                        [-aux_column_width, -.5, seperator_distance],
                        [aux_column_width, -.5, seperator_distance],
                        [-aux_column_width, -hd_center, hd_center],
                        [aux_column_width, -hd_center, hd_center],
                        [-aux_z_max, -aux_z_max, hd_center],
                        [aux_z_max, -aux_z_max, hd_center],
                        [-aux_column_width, -.5, .5],
                        [aux_column_width, -.5, .5],
                        [-aux_column_width, -seperator_distance, .5],
                        [aux_column_width, -seperator_distance, .5],
                        [-z_max_r, -z_max_r, .5],
                        [z_max_r, -z_max_r, .5],
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[1, 2, 2], control_points=y_min_z_max)
        )

        y_max_z_max = np.array(
                [
                        [-center_r, center_r, center_r],
                        [center_r, center_r, center_r],
                        [-aux_y_max, hd_center, aux_y_max],
                        [aux_y_max, hd_center, aux_y_max],
                        [-y_max_r, .5, y_max_r], [y_max_r, .5, y_max_r],
                        [-aux_z_max, aux_z_max, hd_center],
                        [aux_z_max, aux_z_max, hd_center],
                        [-aux_column_width, hd_center, hd_center],
                        [aux_column_width, hd_center, hd_center],
                        [-aux_column_width, .5, seperator_distance],
                        [aux_column_width, .5, seperator_distance],
                        [-z_max_r, z_max_r, .5], [z_max_r, z_max_r, .5],
                        [-aux_column_width, seperator_distance, .5],
                        [aux_column_width, seperator_distance, .5],
                        [-aux_column_width, .5, .5],
                        [aux_column_width, .5, .5]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(degrees=[1, 2, 2], control_points=y_max_z_max)
        )

        x_min_y_min_z_min = np.array(
                [
                        [-.5, -.5, -.5], [-seperator_distance, -.5, -.5],
                        [-aux_column_width, -.5, -.5],
                        [-.5, -seperator_distance, -.5],
                        [-seperator_distance, -seperator_distance, -.5],
                        [-aux_column_width, -seperator_distance, -.5],
                        [-.5, -aux_column_width, -.5],
                        [-seperator_distance, -aux_column_width, -.5],
                        [-z_min_r, -z_min_r, -.5],
                        [-.5, -.5, -seperator_distance],
                        [-seperator_distance, -.5, -seperator_distance],
                        [-aux_column_width, -.5, -seperator_distance],
                        [-.5, -seperator_distance, -seperator_distance],
                        [-hd_center, -hd_center, -hd_center],
                        [-aux_column_width, -hd_center, -hd_center],
                        [-.5, -aux_column_width, -seperator_distance],
                        [-hd_center, -aux_column_width, -hd_center],
                        [-aux_z_min, -aux_z_min, -hd_center],
                        [-.5, -.5, -aux_column_width],
                        [-seperator_distance, -.5, -aux_column_width],
                        [-y_min_r, -.5, -y_min_r],
                        [-.5, -seperator_distance, -aux_column_width],
                        [-hd_center, -hd_center, -aux_column_width],
                        [-aux_y_min, -hd_center, -aux_y_min],
                        [-.5, -x_min_r, -x_min_r],
                        [-hd_center, -aux_x_min, -aux_x_min],
                        [-center_r, -center_r, -center_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_min_y_min_z_min
                )
        )

        x_max_y_min_z_min = np.array(
                [
                        [aux_column_width, -.5, -.5],
                        [seperator_distance, -.5, -.5], [.5, -.5, -.5],
                        [aux_column_width, -seperator_distance, -.5],
                        [seperator_distance, -seperator_distance, -.5],
                        [.5, -seperator_distance, -.5], [
                                z_min_r, -z_min_r, -.5
                        ], [seperator_distance, -aux_column_width, -.5],
                        [.5, -aux_column_width, -.5],
                        [aux_column_width, -.5, -seperator_distance],
                        [seperator_distance, -.5, -seperator_distance],
                        [.5, -.5, -seperator_distance],
                        [aux_column_width, -hd_center, -hd_center],
                        [hd_center, -hd_center, -hd_center],
                        [.5, -seperator_distance, -seperator_distance],
                        [aux_z_min, -aux_z_min, -hd_center],
                        [hd_center, -aux_column_width, -hd_center],
                        [.5, -aux_column_width, -seperator_distance],
                        [y_min_r, -.5, -y_min_r],
                        [seperator_distance, -.5, -aux_column_width],
                        [.5, -.5, -aux_column_width],
                        [aux_y_min, -hd_center, -aux_y_min],
                        [hd_center, -hd_center, -aux_column_width],
                        [.5, -seperator_distance, -aux_column_width],
                        [center_r, -center_r, -center_r],
                        [hd_center, -aux_x_max, -aux_x_max],
                        [.5, -x_max_r, -x_max_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_max_y_min_z_min
                )
        )

        x_min_y_max_z_min = np.array(
                [
                        [-.5, aux_column_width, -.5],
                        [-seperator_distance, aux_column_width, -.5],
                        [-z_min_r, z_min_r, -.5], [
                                -.5, seperator_distance, -.5
                        ], [-seperator_distance, seperator_distance, -.5],
                        [-aux_column_width, seperator_distance, -.5],
                        [-.5, .5, -.5], [-seperator_distance, .5, -.5],
                        [-aux_column_width, .5, -.5],
                        [-.5, aux_column_width, -seperator_distance],
                        [-hd_center, aux_column_width, -hd_center],
                        [-aux_z_min, aux_z_min, -hd_center],
                        [-.5, seperator_distance, -seperator_distance],
                        [-hd_center, hd_center, -hd_center],
                        [-aux_column_width, hd_center, -hd_center],
                        [-.5, .5, -seperator_distance],
                        [-seperator_distance, .5, -seperator_distance],
                        [-aux_column_width, .5, -seperator_distance],
                        [-.5, x_min_r, -x_min_r],
                        [-hd_center, aux_x_min, -aux_x_min],
                        [-center_r, center_r, -center_r],
                        [-.5, seperator_distance, -aux_column_width],
                        [-hd_center, hd_center, -aux_column_width],
                        [-aux_y_max, hd_center, -aux_y_max],
                        [-.5, .5, -aux_column_width],
                        [-seperator_distance, .5, -aux_column_width],
                        [-y_max_r, .5, -y_max_r]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_min_y_max_z_min
                )
        )

        x_max_y_max_z_min = np.array(
                [
                        [z_min_r, z_min_r, -.5],
                        [seperator_distance, aux_column_width, -.5],
                        [.5, aux_column_width, -.5],
                        [aux_column_width, seperator_distance, -.5],
                        [seperator_distance, seperator_distance, -.5],
                        [.5, seperator_distance, -.5],
                        [aux_column_width, .5, -.5],
                        [seperator_distance, .5, -.5], [.5, .5, -.5],
                        [aux_z_min, aux_z_min, -hd_center],
                        [hd_center, aux_column_width, -hd_center],
                        [.5, aux_column_width, -seperator_distance],
                        [aux_column_width, hd_center, -hd_center],
                        [hd_center, hd_center, -hd_center],
                        [.5, seperator_distance, -seperator_distance],
                        [aux_column_width, .5, -seperator_distance],
                        [seperator_distance, .5, -seperator_distance],
                        [.5, .5, -seperator_distance],
                        [center_r, center_r, -center_r],
                        [hd_center, aux_x_max, -aux_x_max],
                        [.5, x_max_r, -x_max_r],
                        [aux_y_max, hd_center, -aux_y_max],
                        [hd_center, hd_center, -aux_column_width],
                        [.5, seperator_distance, -aux_column_width],
                        [y_max_r, .5, -y_max_r],
                        [seperator_distance, .5, -aux_column_width],
                        [.5, .5, -aux_column_width]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_max_y_max_z_min
                )
        )

        x_min_y_min_z_max = np.array(
                [
                        [-.5, -.5, aux_column_width],
                        [-seperator_distance, -.5, aux_column_width],
                        [-y_min_r, -.5, y_min_r],
                        [-.5, -seperator_distance, aux_column_width],
                        [-hd_center, -hd_center, aux_column_width],
                        [-aux_y_min, -hd_center, aux_y_min],
                        [-.5, -x_min_r, x_min_r],
                        [-hd_center, -aux_x_min, aux_x_min],
                        [-center_r, -center_r, center_r],
                        [-.5, -.5, seperator_distance],
                        [-seperator_distance, -.5, seperator_distance],
                        [-aux_column_width, -.5, seperator_distance],
                        [-.5, -seperator_distance, seperator_distance],
                        [-hd_center, -hd_center, hd_center],
                        [-aux_column_width, -hd_center, hd_center],
                        [-.5, -aux_column_width, seperator_distance],
                        [-hd_center, -aux_column_width, hd_center],
                        [-aux_z_max, -aux_z_max, hd_center], [-.5, -.5, .5],
                        [-seperator_distance, -.5, .5],
                        [-aux_column_width, -.5, .5],
                        [-.5, -seperator_distance, .5],
                        [-seperator_distance, -seperator_distance, .5],
                        [-aux_column_width, -seperator_distance, .5],
                        [-.5, -aux_column_width, .5],
                        [-seperator_distance, -aux_column_width, .5],
                        [-z_max_r, -z_max_r, .5]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_min_y_min_z_max
                )
        )

        x_max_y_min_z_max = np.array(
                [
                        [y_min_r, -.5, y_min_r],
                        [seperator_distance, -.5, aux_column_width],
                        [.5, -.5, aux_column_width],
                        [aux_y_min, -hd_center, aux_y_min],
                        [hd_center, -hd_center, aux_column_width],
                        [.5, -seperator_distance, aux_column_width],
                        [center_r, -center_r, center_r],
                        [hd_center, -aux_x_max, aux_x_max],
                        [.5, -x_max_r, x_max_r],
                        [aux_column_width, -.5, seperator_distance],
                        [seperator_distance, -.5, seperator_distance],
                        [.5, -.5, seperator_distance],
                        [aux_column_width, -hd_center, hd_center],
                        [hd_center, -hd_center, hd_center],
                        [.5, -seperator_distance, seperator_distance],
                        [aux_z_max, -aux_z_max, hd_center],
                        [hd_center, -aux_column_width, hd_center],
                        [.5, -aux_column_width, seperator_distance],
                        [aux_column_width, -.5, .5],
                        [seperator_distance, -.5, .5],
                        [.5, -.5, .5],
                        [aux_column_width, -seperator_distance, .5],
                        [seperator_distance, -seperator_distance, .5],
                        [.5, -seperator_distance, .5],
                        [z_max_r, -z_max_r, .5],
                        [seperator_distance, -aux_column_width, .5],
                        [.5, -aux_column_width, .5],
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_max_y_min_z_max
                )
        )

        x_min_y_max_z_max = np.array(
                [
                        [-.5, x_min_r, x_min_r],
                        [-hd_center, aux_x_min, aux_x_min],
                        [-center_r, center_r, center_r],
                        [-.5, seperator_distance, aux_column_width],
                        [-hd_center, hd_center, aux_column_width],
                        [-aux_y_max, hd_center, aux_y_max],
                        [-.5, .5, aux_column_width],
                        [-seperator_distance, .5, aux_column_width],
                        [-y_max_r, .5, y_max_r],
                        [-.5, aux_column_width, seperator_distance],
                        [-hd_center, aux_column_width, hd_center],
                        [-aux_z_max, aux_z_max, hd_center],
                        [-.5, seperator_distance, seperator_distance],
                        [-hd_center, hd_center, hd_center],
                        [-aux_column_width, hd_center, hd_center],
                        [-.5, .5, seperator_distance],
                        [-seperator_distance, .5, seperator_distance],
                        [-aux_column_width, .5, seperator_distance],
                        [-.5, aux_column_width, .5],
                        [-seperator_distance, aux_column_width, .5],
                        [-z_max_r, z_max_r, .5], [-.5, seperator_distance, .5],
                        [-seperator_distance, seperator_distance, .5],
                        [-aux_column_width, seperator_distance, .5],
                        [-.5, .5, .5], [-seperator_distance, .5, .5],
                        [-aux_column_width, .5, .5]
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_min_y_max_z_max
                )
        )

        x_max_y_max_z_max = np.array(
                [
                        [center_r, center_r, center_r],
                        [hd_center, aux_x_max, aux_x_max],
                        [.5, x_max_r, x_max_r],
                        [aux_y_max, hd_center, aux_y_max],
                        [hd_center, hd_center, aux_column_width],
                        [.5, seperator_distance, aux_column_width],
                        [y_max_r, .5, y_max_r],
                        [seperator_distance, .5, aux_column_width],
                        [.5, .5, aux_column_width],
                        [aux_z_max, aux_z_max, hd_center],
                        [hd_center, aux_column_width, hd_center],
                        [.5, aux_column_width, seperator_distance],
                        [aux_column_width, hd_center, hd_center],
                        [hd_center, hd_center, hd_center],
                        [.5, seperator_distance, seperator_distance],
                        [aux_column_width, .5, seperator_distance],
                        [seperator_distance, .5, seperator_distance],
                        [.5, .5, seperator_distance],
                        [z_max_r, z_max_r, .5],
                        [seperator_distance, aux_column_width, .5],
                        [.5, aux_column_width, .5],
                        [aux_column_width, seperator_distance, .5],
                        [seperator_distance, seperator_distance, .5],
                        [.5, seperator_distance, .5],
                        [aux_column_width, .5, .5],
                        [seperator_distance, .5, .5],
                        [.5, .5, .5],
                ]
        ) + np.array([.5, .5, .5])

        spline_list.append(
                base.Bezier(
                        degrees=[2, 2, 2], control_points=x_max_y_max_z_max
                )
        )

        return spline_list



"""gustaf/gustaf/vertices.py.

Vertices. Base of all "Mesh" geometries.
"""

import copy

import numpy as np

from gustaf import settings
from gustaf import utils
from gustaf import show
from gustaf import helpers
from gustaf._base import GustafBase


class Vertices(GustafBase):

    kind = "vertex"

    __slots__ = [
            "_vertices",
            "_const_vertices",
            "_computed",
            "vis_dict",
            "vertexdata",
    ]

    def __init__(
            self,
            vertices=None,
    ):
        """Vertices. It has vertices.

        Parameters
        -----------
        vertices: (n, d) np.ndarray

        Returns
        --------
        None
        """
        if vertices is not None:
            self.vertices = vertices

        self._computed = helpers.data.ComputedMeshData(self)

        self.vis_dict = dict()
        self.vertexdata = dict()

    @property
    def vertices(self):
        """Returns vertices.

        Parameters
        -----------
        None

        Returns
        --------
        vertices: (n, d) np.ndarray
        """
        self._logd("returning vertices")
        return self._vertices

    @vertices.setter
    def vertices(self, vs):
        """Vertices setter. This will saved as a tracked array. This tracked
        array is very sensitive and if we do anything with it that may hint an
        inplace operation, it will be marked as modified. This includes copying
        and slicing. If you know you aren't going to modify the array, please
        consider using `const_vertices`. Somewhat c-style hint in naming.

        Parameters
        -----------
        vs: (n, d) np.ndarray

        Returns
        --------
        None
        """
        self._logd("setting vertices")

        # shape check
        utils.arr.is_shape(vs, (-1, -1), strict=True)

        self._vertices = helpers.data.make_tracked_array(
                vs, settings.FLOAT_DTYPE
        )
        # exact same, but not tracked.
        self._const_vertices = self._vertices.view()
        self._const_vertices.flags.writeable = False

    @property
    def const_vertices(self):
        """Returns non-mutable view of `vertices`. Naming inspired by c/cpp
        sessions.

        Parameters
        -----------
        None

        Returns
        --------
        None
        """
        self._logd("returning const_vertices")
        return self._const_vertices

    @property
    def whatami(self, ):
        """Answers deep philosophical question: "what am i"?

        Parameters
        ----------
        None

        Returns
        --------
        whatami: str
          vertices
        """
        return "vertices"

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def unique_vertices(self, tolerance=None, **kwargs):
        """Returns a namedtuple that holds unique vertices info. Unique here
        means "close-enough-within-tolerance".

        Parameters
        -----------
        tolerance: float
          (Optional) Default is settings.TOLERANCE
        recompute: bool
          Only applicable as keyword argument. Force re-computes.

        Returns
        --------
        unique_vertices_info: Unique2DFloats
          namedtuple with `values`, `ids`, `inverse`, `intersection`.
        """
        self._logd("computing unique vertices")
        if tolerance is None:
            tolerance = settings.TOLERANCE

        values, ids, inverse, intersection = utils.arr.close_rows(
                self.const_vertices, tolerance=tolerance
        )

        return helpers.data.Unique2DFloats(
                values,
                ids,
                inverse,
                intersection,
        )

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def bounds(self):
        """Returns bounds of the vertices. Bounds means AABB of the geometry.

        Parameters
        -----------
        None

        Returns
        --------
        bounds: (d,) np.ndarray
        """
        self._logd("computing bounds")
        return utils.arr.bounds(self.const_vertices)

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def bounds_diagonal(self):
        """Returns diagonal vector of the bounding box.

        Parameters
        -----------
        None

        Returns
        --------
        bounds_digonal: (d,) np.ndarray
          same as `bounds[1] - bounds[0]`
        """
        self._logd("computing bounds_diagonal")
        bounds = self.bounds()
        return bounds[1] - bounds[0]

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def bounds_diagonal_norm(self):
        """Returns norm of bounds diagonal.

        Parameters
        -----------
        None

        Returns
        --------
        bounds_diagonal_norm: float
        """
        self._logd("computing bounds_diagonal_norm")
        return float(sum(self.bounds_diagonal()**2)**.5)

    def update_vertices(self, mask, inverse=None):
        """Update vertices with a mask. In other words, keeps only masked
        vertices. Adapted from `github.com/mikedh/trimesh`. Updates
        connectivity accordingly too.

        Parameters
        -----------
        mask: (n,) bool or int
        inverse: (len(self.vertices),) int

        Returns
        --------
        updated_self: type(self)
        """
        vertices = self.const_vertices.copy()

        # make mask numpy array
        mask = np.asarray(mask)

        if ((mask.dtype.name == "bool" and mask.all()) or len(mask) == 0):
            return self

        # create inverse mask if not passed
        check_neg = False
        if inverse is None and self.kind != "vertex":
            inverse = np.full(len(vertices), -11, dtype=settings.INT_DTYPE)
            check_neg = True
            if mask.dtype.kind == "b":
                inverse[mask] = np.arange(mask.sum())
            elif mask.dtype.kind == "i":
                inverse[mask] = np.arange(len(mask))
            else:
                inverse = None

        # re-index elements from inverse
        # TODO: Here could be a good place to preserve BCs.
        elements = None
        if inverse is not None and self.kind != "vertex":
            elements = self.const_elements.copy()
            elements = inverse[elements.reshape(-1)].reshape(
                    (-1, elements.shape[1])
            )
            # remove all the elements that's not part of inverse
            if check_neg:
                emask = (elements > -1).all(axis=1)
                elements = elements[emask]

        # apply mask
        vertices = vertices[mask]

        def update_vertexdata(obj, m, vertex_data=None):
            """apply mask to vertex data if there's any."""
            newdata = dict()
            if vertex_data is None:
                vertex_data = obj.vertexdata

            for key, values in vertex_data.items():
                newdata[key] = values[m]

            obj.vertexdata = newdata

            return obj

        # update
        self.vertices = vertices
        if elements is not None:
            self.elements = elements

        update_vertexdata(self, mask)

        return self

    def select_vertices(self, ranges):
        """Returns vertices inside the given range.

        Parameters
        -----------
        ranges: (d, 2) array-like
          Takes None.

        Returns
        --------
        ids: (n,) np.ndarray
        """
        return utils.arr.select_with_ranges(self.vertices, ranges)

    def remove_vertices(self, ids):
        """Removes vertices with given vertex ids.

        Parameters
        -----------
        ids: (n,) np.ndarray

        Returns
        --------
        new_self: type(self)
        """
        mask = np.ones(len(self.vertices), dtype=bool)
        mask[ids] = False

        return self.update_vertices(mask, )

    def merge_vertices(self, tolerance=None):
        """Based on unique vertices, merge vertices if it is mergeable.

        Parameters
        -----------
        tolerance: float
          Default is settings.TOLERANCE

        Returns
        --------
        merged_self: type(self)
        """
        unique_vs = self.unique_vertices()

        self._logd("number of vertices")
        self._logd(f"  before merge: {len(self.vertices)}")
        self._logd(f"  after merge: {len(unique_vs.ids)}")

        return self.update_vertices(
                mask=unique_vs.ids,
                inverse=unique_vs.inverse,
        )

    def showable(self, **kwargs):
        """Returns showable object, meaning object of visualization backend.

        Parameters
        -----------
        **kwargs:

        Returns
        --------
        showable: obj
          Obj of `gustaf.settings.VISUALIZATION_BACKEND`
        """
        return show.make_showable(self, **kwargs)

    def show(self, **kwargs):
        """Show current object using visualization backend.

        Parameters
        -----------
        **kwargs:


        Returns
        --------
        None
        """
        return show.show(self, **kwargs)

    def copy(self):
        """Returns deepcopy of self.

        Parameters
        -----------
        None

        Returns
        --------
        selfcopy: type(self)
        """
        # all attributes are deepcopy-able
        return copy.deepcopy(self)

    @classmethod
    def concat(cls, *instances):
        """Sequentially put them together to make one object.

        Parameters
        -----------
        *instances: List[type(cls)]
          Allows one iterable object also.

        Returns
        --------
        one_instance: type(cls)
        """

        def is_concatable(inst):
            """Return true, if it is same as type(cls)"""
            if isinstance(inst, cls):
                return True
            else:
                return False

        # If only one instance is given and it is iterable, adjust
        # so that we will just iterate that.
        if (
                len(instances) == 1 and not isinstance(instances[0], str)
                and hasattr(instances[0], "__iter__")
        ):
            instances = instances[0]

        vertices = []
        haselem = cls.kind != "vertex"
        if haselem:
            elements = []

        # check if everything is "concatable".
        for ins in instances:
            if not is_concatable(ins):
                raise TypeError(
                        "Can't concat. One of the instances is not "
                        f"`{cls.__name__}`."
                )

            # make sure each element index starts from 0 & end at len(vertices)
            tmp_ins = ins.copy().remove_unreferenced_vertices()

            vertices.append(tmp_ins.vertices.copy())

            if haselem:
                if len(elements) == 0:
                    elements.append(tmp_ins.elements.copy())
                    e_offset = elements[-1].max() + 1

                else:
                    elements.append(
                            # copy is not necessary here,
                            tmp_ins.elements + e_offset
                    )
                    e_offset = elements[-1].max() + 1

        if haselem:
            return cls(
                    vertices=np.vstack(vertices),
                    elements=np.vstack(elements),
            )

        else:
            return Vertices(vertices=np.vstack(vertices), )

    def __add__(self, to_add):
        """Concat in form of +.

        Parameters
        -----------
        to_add: type(self)

        Returns
        --------
        added: type(self)
        """
        return type(self).concat(self, to_add)



"""gustaf/gustaf/helpers/raise_if.py.

Collection of wrapper functions/classes that raises Error with certain
behavior
"""
from typing import Any, Optional


def invalid_inherited_attr(func, qualname, property_=False):
    """Returns a function that would behave the same as given function, but
    would raise AttributeError. This needs to be defined in class level.

    Parameters
    ----------
    func: (function)
        _description_
    qualname: (class)
        _description_
    property_: (bool, optional)
        is this function a property?. Defaults to False.

    Returns
    -------
    raiser: function
        behaves same as func if `property_` is correctly defined
    """

    def raiser(self):
        raise AttributeError(
                f"{func.__name__} is not supported from {qualname} "
                "and its subclasses thereof."
        )

    if property_:
        return property(raiser)

    else:
        return raiser


class ModuleImportRaiser:
    """Mock imports optional modules if they are not installed.

    Class used to have better import error handling in the case that a
    package package is not installed. This is necessary due to that some
    packages are not a dependency of `gustaf`, but some parts require
    them to function. Examples are `splinepy` and `vedo`.
    """

    def __init__(self, lib_name: str, error_message: Optional[str] = None):
        original_message = ""
        if error_message is not None:
            original_message = f"\nOriginal error message - {error_message}"
        self._message = str(
                f"Cannot load {lib_name} package, on which requested "
                "functionality depends. "
                "Please refer to the installation instructions "
                "[tataratat.github.io/gustaf] for more information."
                f"{original_message}"
        )

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """Is called when the object is called by object().

        Will notify the user, that the functionality is not accessible
        and how to proceed to access the functionality.
        """
        raise ImportError(self._message)

    def __getattr__(self, __name: str) -> Any:
        """Is called when any attribute of the object is accessed by
        object.attr.

        Will notify the user, that the functionality is not accessible
        and how to proceed to access the functionality.
        """
        if __name == "_ModuleImportRaiser__message":
            return object.__getattr__(self, __name[-8:])
        else:
            raise ImportError(self._message)

    def __setattr__(self, __name: str, __value: Any) -> None:
        """
        Is called when any attribute of the object is set by object.attr = new.
        Will notify the user, that the functionality is not accessible and how
        to proceed to access the functionality.
        """
        if __name == "_message":
            object.__setattr__(self, __name, __value)
        else:
            raise ImportError(self._message)

    def __getitem__(self, key):
        """Is called when the object is subscripted object[x].

        Will notify the user, that the functionality is not accessible
        and how to proceed to access the functionality.
        """
        raise ImportError(self._message)
